/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"


/*	search_interface

	Function to search the InterfacesList. This is a very slow
	function. It worst case performance is O(n) where 'n' is the
	size of the list.

	It returns the pointer or a NULL-pointer if not found.
*/
void *
search_interface(const int debuglvl, const char *name, const d_list *d_list)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;

	/* safety check */
	if(!name || !d_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "looking for interface '%s'.", name);

	/*
		dont bother searching a empty list
	*/
	if(d_list->len == 0)
		return(NULL);

	/*
		loop trough the list and compare the names
	*/
	for(d_node = d_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(NULL);
		}

		if(strcmp(iface_ptr->name, name) == 0)
		{
			/* Found! */
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "Interface '%s' found!", name);

			/* return the pointer we found */
			return(iface_ptr);
		}
	}

	/* if we get here, the interface was not found, so return NULL */
	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "interface '%s' not found.", name);

	return(NULL);
}


/*- print_list -

	Prints the interface list to stdout
*/
void
interfaces_print_list(const d_list *d_list)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;

	if(!d_list)
		return;

	if(d_list->len > 0)
		fprintf(stdout, "list len is %d\n", d_list->len);
	else
	{
		fprintf(stdout, "list is empty.\n");
		return;
	}

	for(d_node = d_list->top; d_node ; d_node = d_node->next)
	{
		iface_ptr = d_node->data;
		fprintf(stdout, "iface: %s, status: %d, active: %d, device: %s, ipaddress: %s\n", iface_ptr->name, iface_ptr->status, iface_ptr->active, iface_ptr->device, iface_ptr->ipv4.ipaddress);
	}

	return;
}


/*	read_interface_info

	Gets the info from the backend:
	Active
	Interface
	Ipaddress

	Returncodes:
	-1: error
	 0: ok

	This function will not fail because of missing or malformed data in the
	backend. It will issue an error but set the interface to inactive.
*/
int
read_interface_info(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	int	result = 0,
		ipresult = 0;
	char	ipaddress[16] = "",
		yesno[4] = "";


	/* safety first */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "start: name: %s", iface_ptr->name);


	/* check if the interface is active */
	result = check_active(debuglvl, iface_ptr->name, TYPE_INTERFACE);
	if(result == 1)
	{
		iface_ptr->active = 1;
	}
	else if(result == 0)
	{
		(void)vrprint.info("Info", "Note: interface '%s' is not active.", iface_ptr->name);
		iface_ptr->active = 0;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "check_active failed for interface %s. Please check the configuration of the interface (in: %s).", iface_ptr->name, __FUNCTION__);
		iface_ptr->active = 0;
	}


	/* ask the backend about the interface of this interface. Get it?
	
		TODO rename this to DEVICE
	*/
	result = af->ask(debuglvl, ifac_backend, iface_ptr->name, "INTERFACE", iface_ptr->device, sizeof(iface_ptr->device), TYPE_INTERFACE, 0);
	if(result == 1)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "interface: %s.", iface_ptr->device);
	}
	else if(result == 0)
	{
		/* if the interface is undefined, issue a warning and set inactive */
		(void)vrprint.warning("Warning", "No INTERFACE defined for interface '%s', deactivating interface.", iface_ptr->name);
		iface_ptr->active = 0;
	}
	else
	{
		/* if we encounter an error, we set inactive */
		(void)vrprint.error(-1, "Error", "looking up INTERFACE for interface '%s' failed (in: %s).", iface_ptr->name, __FUNCTION__);
		return(-1);
	}


	/* ask the backend about the possible virtualness of the device */
	result = af->ask(debuglvl, ifac_backend, iface_ptr->name, "VIRTUAL", yesno, sizeof(yesno), TYPE_INTERFACE, 0);
	if(result == 1)
	{
		if(strcasecmp(yesno, "yes") == 0)
			iface_ptr->device_virtual = 1;
		else if(strcasecmp(yesno, "no") == 0)
			iface_ptr->device_virtual = 0;
		else
		{
			(void)vrprint.warning("Warning", "Unexpected answer '%s' for VIRTUAL for interface '%s', assuming not virtual.", yesno, iface_ptr->name);
			iface_ptr->device_virtual = 0;
		}
	}
	else if(result == 0)
	{
		/* if the interface is undefined, issue a warning and set inactive */
		if(debuglvl >= LOW)
			(void)vrprint.debug(__FUNCTION__, "No VIRTUAL defined for interface '%s', assuming not virtual.", iface_ptr->name);

		iface_ptr->device_virtual = 0;
	}
	else
	{
		/* if we encounter an error, we set inactive */
		(void)vrprint.error(-1, "Error", "looking up VIRTUAL for interface '%s' failed (in: %s).", iface_ptr->name, __FUNCTION__);
		return(-1);
	}


	/* ask the ipaddress of this interface */
	result = af->ask(debuglvl, ifac_backend, iface_ptr->name, "IPADDRESS", iface_ptr->ipv4.ipaddress, sizeof(iface_ptr->ipv4.ipaddress), TYPE_INTERFACE, 0);
	if(result == 1)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "ipaddress: %s.", iface_ptr->ipv4.ipaddress);

		/* check if ip is dynamic */
		if(strcmp(iface_ptr->ipv4.ipaddress, "dynamic") == 0)
		{
			iface_ptr->dynamic = 1;

			/* now try to get the dynamic ipaddress */
			ipresult = get_dynamic_ip(debuglvl, iface_ptr->device, iface_ptr->ipv4.ipaddress, sizeof(iface_ptr->ipv4.ipaddress));
			if(ipresult == 0)
			{
				/* interface is down */
				(void)vrprint.debug(__FUNCTION__, "getting ipaddress for dynamic interface '%s' (%s) failed because interface is down. Deactivating.", iface_ptr->name, iface_ptr->device);
				iface_ptr->active = 0;
			}
			else if(ipresult < 0)
			{
				/* something went wrong */
				(void)vrprint.warning("Warning", "getting the dynamic ipaddress for interface '%s' (%s) failed. Deactivating interface.", iface_ptr->name, iface_ptr->device);
				iface_ptr->active = 0;
			}
		}

		/* check the ip */
		if(strcmp(iface_ptr->ipv4.ipaddress, "dynamic") != 0)
		{
			if(check_ipv4address(debuglvl, NULL, NULL, iface_ptr->ipv4.ipaddress, 0) != 1)
			{
				(void)vrprint.warning("Warning", "the ipaddress '%s' of interface '%s' (%s) is invalid. Deactivating interface.", iface_ptr->ipv4.ipaddress, iface_ptr->name, iface_ptr->device);
				iface_ptr->active = 0;
			}
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "No IPADDRESS defined for interface '%s', deactivating interface.", iface_ptr->name);
		iface_ptr->active = 0;
	}
	else
	{
		/* if we encounter an error, we set inactive */
		(void)vrprint.error(-1, "Error", "read_interface_info: ask_backend returned error. Looking up ipaddress for '%s' failed.", iface_ptr->name);
		return(-1);
	}


	/* get the protect rules */
	if(interfaces_get_protectrules(debuglvl, iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "getting protect rules for interface '%s' failed.", iface_ptr->name);
		return(-1);
	}


	/* now check if the interface is currently up */
	result = get_iface_stats(iface_ptr->device, NULL, NULL, NULL, NULL);
	if(result == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "interface '%s' is up.", iface_ptr->name);

		iface_ptr->up = 1;
	}
	else if(result == 1)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "interface '%s' is down.", iface_ptr->name);

		iface_ptr->up = 0;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "get_iface_stats returned error.");
		return(-1);
	}


	/* if the interface is up check the ipaddress with the ipaddress we know */
	if(iface_ptr->up && iface_ptr->active)
	{
		if(get_dynamic_ip(debuglvl, iface_ptr->device, ipaddress, sizeof(ipaddress)) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "checking the ipaddress of the interface '%s' failed (in: %s).", iface_ptr->name, __FUNCTION__);
			return(-1);
		}

		if(strcmp(ipaddress, iface_ptr->ipv4.ipaddress) != 0)
		{
			(void)vrprint.warning("Warning", "the ipaddress '%s' of interface '%s' (%s) does not match the ipaddress of the actual interface (%s). Deactivating interface.", iface_ptr->ipv4.ipaddress, iface_ptr->name, iface_ptr->device, ipaddress);
			iface_ptr->active = 0;
		}
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "end: succes. name: %s.", iface_ptr->name);

	return(0);
}


/*	insert_interface

	Inserts the interface 'name' into the linked-list.

	Returncodes:
		-1: error
		 0: succes
		 1: interface failed, maybe it is inactive
*/
int
insert_interface(const int debuglvl, d_list *d_list, char *name)
{
	struct InterfaceData_	*iface_ptr = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "start: name: %s.", name);


	/* safety */
	if(!name || !d_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* claiming the memory we need */
	if(!(iface_ptr = interface_malloc(debuglvl)))
	{
		(void)vrprint.error(-1, "Error", "claiming the memory for the interface %s failed (in: insert_interface).", name);
		return(-1);
	}


	/* set the name in the new structure */
	if(strlcpy(iface_ptr->name, name, sizeof(iface_ptr->name)) > sizeof(iface_ptr->name))
	{
		(void)vrprint.error(-1, "Internal Error", "buffer overflow! (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* call read_interface_info. here the info is read. */
	if(read_interface_info(debuglvl, iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "getting info for interface %s failed (in: insert_interface).", name);
		return(-1);
	}


	/* insert into the list */
	if(d_list_append(debuglvl, d_list, iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "inserting interface %s into the list failed (in: insert_interface).", name);

		free(iface_ptr);
		return(-1);
	}


	/* update status */
	iface_ptr->status = ST_ADDED;


	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "end: succes.");

	return(0);
}


/*	init_interfaces

	Loads all interfaces in memory.

	Returncodes:
		 0: succes
		 1: succes with one or more interfaces entries failed
		-1: error
*/
int
init_interfaces(const int debuglvl, d_list *interfaces_list)
{
	int	result = 0,
		counter = 0,
		zonetype = 0;
	char	ifacname[MAX_INTERFACE] = "";

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "start");


	/* safety */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* setup the list */
	if(d_list_setup(debuglvl, interfaces_list, NULL) < 0)
		return(-1);


	/* get the list from the backend */
	while(af->list(debuglvl, ifac_backend, ifacname, &zonetype, CAT_INTERFACES) != NULL)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "loading interface %s", ifacname);

		result = insert_interface(debuglvl, interfaces_list, ifacname);
		if(result < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "inserting interface %s into list failed (in: %s).", ifacname, __FUNCTION__);
			return(-1);
		}
		else
		{
			counter++;

			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "loading interface succes: '%s'.", ifacname);
		}
	}

	(void)vrprint.info("Info", "%d interfaces loaded.", counter);

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "end");

	return(0);
}


int
new_interface(const int debuglvl, char *iface_name, d_list *d_list)
{
	int			result = 0;
	struct InterfaceData_	*iface_ptr = NULL;


	/* safety */
	if(!iface_name || !d_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* claim memory */
	if(!(iface_ptr = interface_malloc(debuglvl)))
	{
		(void)vrprint.error(-1, "Error", "claiming the memory for the interface %s failed (in: %s).", iface_name, __FUNCTION__);
		return(-1);
	}

	/* copy name */
	if(strlcpy(iface_ptr->name, iface_name, sizeof(iface_ptr->name)) >= sizeof(iface_ptr->name))
	{
		(void)vrprint.error(-1, "Internal Error", "buffer overflow! (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* insert into the list */
	if(d_list_append(debuglvl, d_list, iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "inserting interface '%s' into the interface-list failed (in: new_interface).", iface_name);
		return(-1);
	}


	/* add to the backend */
	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "calling af->add for '%s'.", iface_name);

	result = af->add(debuglvl, ifac_backend, iface_name, TYPE_INTERFACE);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "adding interface '%s' to the backend failed (in: new_interface).", iface_name);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "calling af->add for '%s' succes.", iface_name);

	/* set active */
	result = af->tell(debuglvl, ifac_backend, iface_ptr->name, "ACTIVE", iface_ptr->active ? "Yes" : "No", 1, TYPE_INTERFACE);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "saving to backend failed (in: new_interface).");
		return(-1);
	}

	/* set virtual */
	result = af->tell(debuglvl, ifac_backend, iface_ptr->name, "VIRTUAL", iface_ptr->device_virtual ? "Yes" : "No", 1, TYPE_INTERFACE);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "saving to backend failed (in: new_interface).");
		return(-1);
	}

	return(0);
}


/*	delete_interface

	Deletes an interface from the list, from memory and from
	the backend.

	Returncodes:
		 0: ok
		-1: error
*/
int
delete_interface(const int debuglvl, char *iface_name, d_list *d_list)
{
	struct InterfaceData_	*iface_ptr = NULL;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!iface_name || !d_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* first search the interface in the list */
	if(!(iface_ptr = search_interface(debuglvl, iface_name, d_list)))
	{
		(void)vrprint.error(-1, "Error", "interface '%s' not found in memory (in: %s).", iface_name, __FUNCTION__);
		return(-1);
	}
	iface_ptr = NULL;


	/* remove the interface from the backend */
	if(af->del(debuglvl, ifac_backend, iface_name, TYPE_INTERFACE, 1) < 0)
	{
		(void)vrprint.error(-1, "Error", "removing interface from backend failed (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* now search the interface again to remove it */
	for(d_node = d_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		if(strcmp(iface_name, iface_ptr->name) == 0)
		{
			/*	this is the interface

				now remove it from the list
			*/
			if(d_list_remove_node(debuglvl, d_list, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "failed to delete the interface '%s' from the list (in: %s).", iface_name, __FUNCTION__);
				return(-1);
			}

			/* finally free the memory */
			free(iface_ptr);

			return(0);
		}
	}


	/* if we get here the interface was not found in the list */
	return(-1);
}


/*	ins_iface_into_zonelist

	load the insertfaces into the zoneslist

	Returncodes
		 0: ok
		-1: error
*/
int
ins_iface_into_zonelist(const int debuglvl, d_list *ifacelist, d_list *zonelist)
{
	struct InterfaceData_	*iface_ptr = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	d_list_node		*iface_node = NULL;
	char			name[MAX_INTERFACE + 8 + 2 + 1]; // 32 max iface length, 8 firewall, 2 () and 1 \0

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "start.");


	/* safety check */
	if(!ifacelist || !zonelist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* dont bother an empty interface list */
	if(ifacelist->len == 0)
		return(0);


	/* loop trough the interface list */
	for(iface_node = ifacelist->top; iface_node; iface_node = iface_node->next)
	{
		if(!(iface_ptr = iface_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/*
			we dont care about an interface without an ipaddress
		*/
		if(strcmp(iface_ptr->ipv4.ipaddress, "") != 0)
		{
			/*
				pretty name
			*/
			if(snprintf(name, sizeof(name), "firewall(%s)", iface_ptr->name) >= sizeof(name))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow! (in: %s).", __FUNCTION__);
				return(-1);
			}

			/*
				alloc mem for the temp zone
			*/
			if(!(zone_ptr = zone_malloc(debuglvl)))
			{
				(void)vrprint.error(-1, "Internal Error", "zone_malloc failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			/*
				copy the name
			*/
			if(strlcpy(zone_ptr->name, name, sizeof(zone_ptr->name)) > sizeof(zone_ptr->name))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow! (in: %s).", __FUNCTION__);
				return(-1);
			}

			/*
				copy the ipaddress
			*/
			if(strlcpy(zone_ptr->ipv4.ipaddress, iface_ptr->ipv4.ipaddress, sizeof(zone_ptr->ipv4.ipaddress)) > sizeof(zone_ptr->ipv4.ipaddress))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow! (in: %s).", __FUNCTION__);
				return(-1);
			}

			/*
				set the type to firewall, so we can recognize the interface in an easy way
			*/
			zone_ptr->type = TYPE_FIREWALL;

			/*
				append to the zoneslist
			*/
			if(d_list_append(debuglvl, zonelist, zone_ptr) < 0)
			{
				(void)vrprint.error(-1, "Error", "appending zonedata to the list failed (in: %s).", __FUNCTION__);

				free(zone_ptr);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "inserted '%s' into zonelist.", zone_ptr->name);

		}
	}

	return(0);
}


/*	rem_iface_from_zonelist

	Removes all zones with type TYPE_FIREWALL from the zoneslist.
	This normally are interfaces and network broadcast addresses
	which were included in this list by ins_iface_into_zonelist.

	Returncodes:
		 0: ok
		-1: error
*/
int
rem_iface_from_zonelist(const int debuglvl, d_list *zonelist)
{
	struct ZoneData_	*zone_ptr = NULL;
	d_list_node		*d_node = NULL,
				*next_node = NULL;
	int			i = 0;

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "start.");

	/*
		safety
	*/
	if(!zonelist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: rem_iface_from_zonelist).");
		return(-1);
	}

	for(d_node = zonelist->top; d_node; d_node = next_node)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: rem_iface_from_zonelist).");
			return(-1);
		}

		/*
			we use next_node here because when a d_node is
			removed from the list it is also free'd from
			memory. So we have to determine the next
			node before that happens.
		*/
		next_node = d_node->next;

		if(zone_ptr->type == TYPE_FIREWALL)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "type: TYPE_FIREWALL '%s'.", zone_ptr->name);

			/*
				remove the node from the list
			*/
			if(d_list_remove_node(debuglvl, zonelist, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "removing from list failed (in: rem_iface_from_zonelist).");
				return(-1);
			}

			/*
				free the memory, but only if the remove function
				in the list is NULL. Otherwise it is already free'd
				by d_list_remove_node.
			*/
			if(zonelist->remove == NULL)
				free(zone_ptr);

			i++;
		}
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "%d interfaces/broadcasts removed.", i);

	return(0);
}


/*	get_iface_stats

	Gets information about an interface from /proc/net/dev. It can also be used to check if an interface
	is up.

	Returncodes:
		 0: ok
		-1: error
		 1: int not found
*/
int
get_iface_stats(const char *iface_name, long long *recv_bytes, int *recv_packets, long long *trans_bytes, int *trans_packets)
{
	char	proc_net_dev[] = "/proc/net/dev",
		line[512] = "",

		/*
			NOTE: if you change the length of the interface, also change it in
			sscanf!!!!
		*/
		interface[32],
		recv_byte_str[32];

	int	copy_bytes = 0,
		i = 0,
		k = 0,
		found = 0;	// indicates that the interface was found

	FILE	*fp = NULL;

	struct
	{
		long long bytes; // a long long because otherwise it would max handle 2gb TODO: long should be big enough
		int packets;
		int errors;
		int drop;
		int fifo;
		int frame;
		int comp;
		int multi;
	}	recv  = {0, 0, 0, 0, 0, 0, 0, 0},
		trans = {0, 0, 0, 0, 0, 0, 0, 0};


	/* first reset */
	if(recv_bytes)
		*recv_bytes = 0;
	if(trans_bytes)
		*trans_bytes = 0;

	/* open the proc entry */
	if(!(fp = fopen(proc_net_dev, "r")))
	{
		(void)vrprint.error(-1, "Error", "unable to open '%s': %s (in: %s).", proc_net_dev, strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* loop trough the file */
	while(fgets(line, sizeof(line), fp) != NULL)
	{
		/*	first scan only the first string, here you can see why (from the file):
			lo: 3335005   17735 ...
			eth0:1055472756 4679465 ...

			notice that with eth0 there is no space between the semicolon and the number.
			Thats where we test for.
		*/
		sscanf(line, "%s", interface);
		if(strncmp(interface, iface_name, strlen(iface_name)) == 0)
		{
			found = 1;

			/*
				if only want to know if the int is up break out now
			*/
			if(!recv_bytes && !trans_bytes)
				break;

			/*
				if we have an semicolon at the end
			*/
			if(interface[strlen(interface)-1] == ':')
			{
				sscanf(line, "%32s %lld %d %d %d %d %d %d %d %lld %d %d %d %d %d %d %d", interface, &recv.bytes, &recv.packets, &recv.errors, &recv.drop, &recv.fifo, &recv.frame, &recv.comp, &recv.multi,
													&trans.bytes, &trans.packets, &trans.errors, &trans.drop, &trans.fifo, &trans.frame, &trans.comp, &trans.multi);
			}
			/*
				else the recv bytes is very big
			*/
			else
			{
				/*
					okay, lets split up
				*/
				for(i=0, k=0; (i < strlen(interface) && k < sizeof(recv_byte_str)); i++)
				{
					if(copy_bytes == 1)
					{
						recv_byte_str[k] = interface[i];
						k++;
					}

					if(interface[i] == ':')
						copy_bytes = 1;
				}
				recv_byte_str[k] = '\0';

				/*
					now convert to long long
				*/
				recv.bytes = atoll(recv_byte_str);

				sscanf(line, "%32s %d %d %d %d %d %d %d %lld %d %d %d %d %d %d %d", interface, &recv.packets, &recv.errors, &recv.drop, &recv.fifo, &recv.frame, &recv.comp, &recv.multi,
												    &trans.bytes, &trans.packets, &trans.errors, &trans.drop, &trans.fifo, &trans.frame, &trans.comp, &trans.multi);
			}

			/*
				pass back to the calling function
			*/
			*recv_bytes = recv.bytes;
			*trans_bytes = trans.bytes;
		}
	}

	if(fclose(fp) < 0)
		return(-1);

	/*
		is not found, return 1
	*/
	if(found == 0)
		return(1);

	return(0);
}


/*	validate_interfacename

	Returncodes:
		0: ok
		-1: error
*/
int
validate_interfacename(const int debuglvl, char *interfacename, regex_t *reg_ex)
{
	/*
		safety
	*/
	if(!interfacename || !reg_ex)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "checking: %s", interfacename);

	/*
		run the regex
	*/
	if(regexec(reg_ex, interfacename, 0, NULL, 0) != 0)
	{
		(void)vrprint.error(-1, "Error", "interfacename '%s' is invalid. A interfacename can contain normal letters and numbers and the underscore (_) and minus (-) characters.", interfacename);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "%s is valid", interfacename);

	return(0);
}


/*	destroy_interfaceslist

*/
void
destroy_interfaceslist(const int debuglvl, d_list *interfaces_list)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;


	/* safety */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return;
	}


	/* first destroy all PortrangeLists */
	for(d_node = interfaces_list->top; d_node ; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return;
		}

		d_list_cleanup(debuglvl, &iface_ptr->ProtectList);

		free(iface_ptr);
		iface_ptr = NULL;
	}

	/* then the list itself */
	d_list_cleanup(debuglvl, interfaces_list);
}

/*	rules_parse_line

	Returncodes:
		0: ok
		-1: error
*/
static int
interfaces_protectrule_parse_line(const int debuglvl, const char *line, struct RuleData_ *rule_ptr)
{
	int	line_pos = 0,	// position in line
		var_pos=0;	// position in varible
	char	against_keyw[32] = "";


	/* safety first */
	if(!line || !rule_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* this should not happen, but it can't hurt to check, right? */
	if(strlen(line) > MAX_RULE_LENGTH)
	{
		(void)vrprint.error(-1, "Internal Error", "rule is too long (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* get the action */
	for(; line_pos < sizeof(rule_ptr->action)-1 && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->action[var_pos] = line[line_pos];
	}
	rule_ptr->action[var_pos] = '\0';

	/* now we analyze the action */
	if(strcasecmp(rule_ptr->action, "protect") != 0)
	{
		(void)vrprint.error(-1, "Error", "expected action protect, got '%s' (in: %s:%d).", rule_ptr->action, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the 'against' */
	for(line_pos++, var_pos = 0; var_pos < sizeof(against_keyw) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		against_keyw[var_pos] = line[line_pos];
	}
	against_keyw[var_pos] = '\0';

	/*
		now check what kind of rule we have
	*/
	if(strcasecmp(against_keyw, "against") != 0)
	{
		(void)vrprint.error(-1, "Error", "expected keyword 'against', got '%s' (in: %s:%d).", against_keyw, __FUNCTION__, __LINE__);
		return(-1);
	}

	/*
		okay, now lets see what kind of danger we are talking about
	*/
	for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.danger) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->danger[var_pos] = line[line_pos];
	}
	rule_ptr->danger[var_pos] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: protect: danger: '%s'", __FUNCTION__, rule_ptr->danger);


	rule_ptr->type = PROT_PROC_INT;
	return(0);
}


int
interfaces_get_protectrules(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	int 			result = 0;
	char			currule[MAX_RULE_LENGTH] = "";
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start **", __FUNCTION__);

	/* safety */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* get all protect rules from the backend */
	while((result = af->ask(debuglvl, ifac_backend, iface_ptr->name, "RULE", currule, sizeof(currule), TYPE_INTERFACE, 1)) == 1)
	{
		/* get mem */
		if(!(rule_ptr = rule_malloc()))
			return(-1);

		/* copy name */
		if(strlcpy(rule_ptr->who, iface_ptr->name, sizeof(rule_ptr->who)) >= sizeof(rule_ptr->who))
		{
			(void)vrprint.error(-1, "Internal Error", "buffer too small (in: %s:%d).", __FUNCTION__, __LINE__);
			free(rule_ptr);
			return(-1);
		}

		(void)vrprint.debug(__FUNCTION__, "currule: '%s'.", currule);

		/* parse the line */
		if(interfaces_protectrule_parse_line(debuglvl, currule, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Error", "parsing protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			free(rule_ptr);
		}
		else
		{
			/* append to list */
			if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) == NULL)
			{
				(void)vrprint.error(-1, "Internal Error", "appending protect rule to list failed (in: %s:%d).", __FUNCTION__, __LINE__);
				free(rule_ptr);
				return(-1);
			}
		}
	}

	for(d_node = iface_ptr->ProtectList.top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			return(-1);
		}

		(void)vrprint.debug(__FUNCTION__, "a: %s, w: %s, d: %s, s: %s.", rule_ptr->action, rule_ptr->who, rule_ptr->danger, rule_ptr->source);
	}

	return(0);
}
