/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <vuurmuur.h>

#define TEXTDIRVERSION_MAJOR	0
#define TEXTDIRVERSION_MINOR	5
#define TEXTDIRVERSION_SUB	54

#define MAX_LINE_LENGTH		512

struct TextdirBackend_
{
	int	version_major;
	int	version_minor;
	int	version_sub;

	/* 0: if backend is closed, 1: open */
	int	backend_open;

	/* not yet used */
	int	writable;

	DIR	*zone_p;
	DIR	*network_p;
	DIR	*host_p;
	DIR	*group_p;

	DIR	*service_p;

	DIR	*interface_p;
	
	FILE	*file;

	char	cur_zone[MAX_ZONE],
		cur_network[MAX_NETWORK],
		cur_host[MAX_HOST];

	char	interface[MAX_INTERFACE];

	char	textdirlocation[512];

	void	*plugin_handle;

	/* regexes for checking the names */
	regex_t		*zonename_reg;
	regex_t		*servicename_reg;
	regex_t		*interfacename_reg;
};


/*	get_filelocation

	get the file location of the 'name' with type 'type'.
	
	Returns NULL on error.
	
	The caller needs to free the memory.
*/
static char
*get_filelocation(const int debuglvl, void *backend, char *name, const int type)
{
	char			hostname[MAX_HOST] = "",
				networkname[MAX_NETWORK] = "",
				zonename[MAX_ZONE] = "";

	char			file_location[512] = "",
				*fileloc_ptr = NULL;

	struct TextdirBackend_	*ptr = NULL;

	/* better safe than sorry */
	if(!backend || !name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}

	if(!(ptr = (struct TextdirBackend_ *)backend))
	{
		(void)vrprint.error(-1, "Internal Error", "backend parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}

	/* check if backend is open */
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "backend not opened yet (in: %s).", __FUNCTION__);
		return(NULL);
	}

	/*
		first zones
	*/
	if(type == TYPE_ZONE || type == TYPE_NETWORK || type == TYPE_GROUP || type == TYPE_HOST)
	{
		if (debuglvl>=HIGH)
			(void)vrprint.debug("Debug", "%s: looking up data from zones.", __FUNCTION__);

		/* validate the name */
		if(validate_zonename(debuglvl, name, 0, zonename, networkname, hostname, ptr->zonename_reg) != 0)
		{
			(void)vrprint.error(-1, "Error", "zonename '%s' is not valid.", name);
			return(NULL);
		}

		/*
			first we determine the zonetype
		*/
		switch(type)
		{
			/* host */
			case TYPE_HOST:

				if(debuglvl >= HIGH)
				{
					(void)vrprint.debug("Debug", "%s: %s is a host.", __FUNCTION__, name);
					(void)vrprint.debug("Debug", "%s: arguments: %s, %s and %s", __FUNCTION__, hostname, networkname, zonename);
				}
				
				/* assemble the filestring, and make sure we dont overflow */
				if(snprintf(file_location, sizeof(file_location), "%s/zones/%s/networks/%s/hosts/%s.host", ptr->textdirlocation, zonename, networkname, hostname) >= sizeof(file_location))
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
					return(NULL);
				}
				
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
				
				break;

			/* group */
			case TYPE_GROUP:
			
				if(debuglvl >= HIGH)
				{
					(void)vrprint.debug("Debug", "%s: %s is a group.", __FUNCTION__, name);
					(void)vrprint.debug("Debug", "%s: arguments: %s, %s and %s", __FUNCTION__, hostname, networkname, zonename);
				}
				
				/* assemble the filestring, and make sure we dont overflow */
				if(snprintf(file_location, sizeof(file_location), "%s/zones/%s/networks/%s/groups/%s.group", ptr->textdirlocation, zonename, networkname, hostname) >= sizeof(file_location))
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
					return(NULL);
				}
				
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
				
				break;

			/* network */
			case TYPE_NETWORK:
				
				if (debuglvl >= HIGH)
				{
					(void)vrprint.debug("Debug", "%s: %s is a network.", __FUNCTION__, name);
					(void)vrprint.debug("Debug", "%s: arguments: %s and %s.", __FUNCTION__, networkname, zonename);
				}
				
				/* assemble the filestring, and make sure we dont overflow */
				if(snprintf(file_location, sizeof(file_location), "%s/zones/%s/networks/%s/network.config", ptr->textdirlocation, zonename, networkname) >= sizeof(file_location))
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
					return(NULL);
				}
				
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
				
				break;

			/* zone */
			case TYPE_ZONE:
				
				if (debuglvl >= HIGH)
				{
					(void)vrprint.debug("Debug", "%s: %s is a zone.", __FUNCTION__, name);
					(void)vrprint.debug("Debug", "%s: arguments: %s.", __FUNCTION__, zonename);
				}
				
				/* assemble the filestring, and make sure we dont overflow */
				if(snprintf(file_location, sizeof(file_location), "%s/zones/%s/zone.config", ptr->textdirlocation, zonename) >= sizeof(file_location))
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
					return(NULL);
				}
				
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
				
				break;
		}

	}

	/*
		services are next
	*/
	else if(type == TYPE_SERVICE || type == TYPE_SERVICEGRP)
	{
		/* validate the name */
		if(validate_servicename(debuglvl, name, ptr->servicename_reg) != 0)
		{
			(void)vrprint.error(-1, "Error", "servicename '%s' is not valid.", name);
			return(NULL);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: looking up data from services, service: %s.", __FUNCTION__, name);
		
		/* assemble the filestring, and make sure we dont overflow */
		if(snprintf(file_location, sizeof(file_location), "%s/services/%s", ptr->textdirlocation, name) >= sizeof(file_location))
		{
			(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
			return(NULL);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
	}

	/*
		interfaces are next
	*/
	else if(type == TYPE_INTERFACE)
	{
		/* validate the name */
		if(validate_interfacename(debuglvl, name, ptr->interfacename_reg) != 0)
		{
			(void)vrprint.error(-1, "Error", "interfacename '%s' is not valid.", name);
			return(NULL);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: looking up data from interfaces, interface: %s.", __FUNCTION__, name);
			
		/* assemble the filestring, and make sure we dont overflow */
		if(snprintf(file_location, sizeof(file_location), "%s/interfaces/%s.conf", ptr->textdirlocation, name) >= sizeof(file_location))
		{
			(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
			return(NULL);
		}
		
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: file_location: %s.", __FUNCTION__, file_location);
	}
	
	/* well, this should not happen, right? */
	else
	{
		(void)vrprint.error(-1, "Internal Error", "unknown type of question '%d' (in: %s).", type, __FUNCTION__);
		return(NULL);
	}

	/* now allocate some memory */
	if(!(fileloc_ptr = malloc(strlen(file_location)+1)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(NULL);
	}

	/* now copy the string to the new mem */
	if(strlcpy(fileloc_ptr, file_location, strlen(file_location)+1) >= strlen(file_location)+1)
	{
		(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
		free(fileloc_ptr);
		return(NULL);
	}

	/* return it! */
	return(fileloc_ptr);
}


/*
	asking from and telling to the backend (TODO: name)
*/
static int
ask_textdir(const int debuglvl, void *backend, char *name, char *question, char *answer, size_t max_answer, int type, int multi)
{
	int			retval = 0;
	char			*file_location = NULL;
	char			line[MAX_LINE_LENGTH],
				variable[64],
				value[512];
	int			i=0,
				j=0,
				k=0,
				line_pos=0,
				val_pos=0;
	char			delt = 'a' - 'A';
	int			line_length=0;
	struct TextdirBackend_	*ptr = NULL;

	/* better safe than sorry */
	if(!backend || !name || !question)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: question: %s, name: %s, multi: %d", __FUNCTION__, question, name, multi);

	if(!(ptr = (struct TextdirBackend_ *)backend))
		return(-1);
	
	/* check if backend is open */
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "backend not opened yet (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* convert question to uppercase: see pp 197 of 'sams teach yourself c in 24 hours' */
	while(question[i])
	{
		if((question[i] >= 'a') && (question[i] <= 'z')) question[i] -= delt;
		++i;
	}

	/* determine the location of the file */
	if(!(file_location = get_filelocation(debuglvl, backend, name, type)))
		return(-1);

	/* check if we are clean */
	if(ptr->file != NULL && multi == 0)
	{
		(void)vrprint.warning("Warning", "the last 'multi' call to '%s' probably failed, because the file is still open when it shouldn't.", __FUNCTION__);
		
		fclose(ptr->file);
		ptr->file = NULL;
	}

	/* now open and read the file, but only if it is not already open */
	if(ptr->file == NULL)
	{
		if(!(ptr->file = vuurmuur_fopen(file_location, "r")))
		{
			(void)vrprint.error(-1, "Error", "Unable to open file '%s'.", file_location);

			free(file_location);
			return(-1);
		}
	}

	/* start (or continue) looping trough the file */
	while(fgets(line, sizeof(line), ptr->file) != NULL)
	{
		line_length=0, k=0; line_pos=0, val_pos=0, j=0;
			
		line_length = strlen(line);
		if(line_length < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "unable to determine the length of 'line' (in: %s).", __FUNCTION__);

			free(file_location);
			fclose(ptr->file);
			ptr->file = NULL;

			return(-1);
		}
		else if(line_length > MAX_LINE_LENGTH)
		{
			(void)vrprint.error(-1, "Error", "line is longer than allowed (line: %d, max: %d) (in: %s).", line_length, MAX_LINE_LENGTH, __FUNCTION__);
			
			free(file_location);
			fclose(ptr->file);
			ptr->file = NULL;

			return(-1);
		}

		/* first check if the line is a comment. */
		if(line_length == 0 || line[0] == '#' || line[0] == ' ' || line[0] == '\0' || line[0] == '\n')
		{
			/* do nothing, its a comment or an empty line. */
		}
		else
		{
			/*	variable
				
				before the '=' we consider variable
			*/
			while((line[k] != '=' && (k < MAX_LINE_LENGTH))) //
			{
				variable[j]=line[k];
				k++;
				j++;
			}
			variable[j]='\0';

			/*
				after that the rest is the value
			*/
			val_pos = 0; line_pos = k + 1;

			while(line[line_pos] != '\0' && line[line_pos] != '\n' && line_length < MAX_LINE_LENGTH && val_pos < max_answer)
			{
				/* if the first character is a '"' we strip it. */
				if((val_pos == 0) && (line[line_pos] == '\"'))
					line_pos++;
				/* otherwise copy the char */
				else
				{
					value[val_pos]=line[line_pos];

					line_pos++;
					val_pos++;
				}
			}

			/* if the last character is a'"' we strip it. */
			if (value[val_pos - 1] == '\"')
				value[val_pos - 1]='\0';
			else
				value[val_pos]='\0';

			/* now see if this was what we were looking for */
			if(strcmp(question, variable) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "%s: question '%s' matched, value: '%s'", __FUNCTION__, question, value);

				if(strlcpy(answer, value, max_answer) >= max_answer)
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt when reading file '%s' (in: %s).", file_location, __FUNCTION__);
					
					free(file_location);
					fclose(ptr->file);
					ptr->file = NULL;
					return(-1);
				}
				
				/* only return when bigger than 0 */
				if(strlen(answer) > 0)
					retval = 1;
					
				/* break out of the loop so when we call multi again we continue where we were */
				break;
			}
		}
	}

	/* cleanup */
	if((multi == 1 && retval != 1) || multi == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "ask_textdir: close the file.");
		
		if(fclose(ptr->file) != 0)
		{
			(void)vrprint.error(-1, "Error", "closing file '%s' failed: %s (in: %s).", file_location, strerror(errno), __FUNCTION__);
			retval = -1;
		}
		ptr->file = NULL;
	}

	/* cleanup filelocation */
	free(file_location);

	if(debuglvl >= HIGH)
	{
		(void)vrprint.debug("Debug", "ask_textdir: ** end **, retval=%d", retval);
		(void)vrprint.debug("Debug", "ask_textdir: at exit: ptr->file: %p (retval: %d).", ptr->file, retval);
	}
	
	return(retval);
}

/*

*/
static int
tell_textdir(int debuglvl, void *backend, char *name, char *question, char *answer, int overwrite, int type)
{
	int			retval=0;
	char			*file_location;
	char			line[512],
				*line_ptr = NULL,
				*tmp_line_ptr = NULL;
	int			i=0,
				found = 0,
				skip = 0;
	int			delta = 'a' - 'A';
	FILE			*fp;
	struct TextdirBackend_	*ptr;
	d_list			storelist;
	d_list_node		*d_node;

	/*
		safety
	*/
	if(!backend || !name || !question || !answer)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: question: %s, answer: %s, name: %s, overwrite: %d, type: %d", __FUNCTION__, question, answer, name, overwrite, type);

	if(!(ptr = (struct TextdirBackend_ *)backend))
	{
		(void)vrprint.error(-1, "Internal Error", "backend parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/*
		check if backend is open
	*/
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "backend not opened yet (in: %s).", __FUNCTION__);
		return(-1);
	}

	/*
		convert question to uppercase
	*/
	while(question[i])
	{
		if((question[i] >= 'a') && (question[i] <= 'z')) question[i] -= delta;
		++i;
	}

	/*
		determine the location of the file
	*/
	if(!(file_location = get_filelocation(debuglvl, backend, name, type)))
		return(-1);

	/*
		first open the file for reading
	*/
	if(!(fp = vuurmuur_fopen(file_location, "r")))
	{
		(void)vrprint.error(-1, "Error", "unable to open file '%s' for reading: %s.", file_location, strerror(errno));

		free(file_location);
		return(-1);
	}

	/*
		initialize the store list
	*/
	if(d_list_setup(debuglvl, &storelist, free) < 0)
	{
		free(file_location);
		fclose(fp);
		return(-1);
	}

	/*
		loop trough the current file
	*/
	while(fgets(line, MAX_LINE_LENGTH, fp) != NULL)
	{
		if(!(line_ptr = malloc(sizeof(line))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);

			/* cleanup */
			d_list_cleanup(debuglvl, &storelist);
			free(file_location);
			fclose(fp);

			return(-1);
		}

		/*
			compare the line
		*/
		if(strncmp(question, line, strlen(question)) == 0)
		{
			if(overwrite && !found)
			{
//TODO
				snprintf(line_ptr, sizeof(line), "%s=\"%s\"\n", question, answer);
				
				found = 1;
			}
			else if(overwrite && found)
			{
				skip = 1;
			}
			else
			{
				if(strlcpy(line_ptr, line, sizeof(line)) >= sizeof(line))
				{
					(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);

					/* cleanup */
					d_list_cleanup(debuglvl, &storelist);
					free(line_ptr);
					free(file_location);
					fclose(fp);
					return(-1);
				}

				found = 1;
			}
		}
		else
		{
			if(strlcpy(line_ptr, line, sizeof(line)) >= sizeof(line))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);

				/* cleanup */
				d_list_cleanup(debuglvl, &storelist);
				free(line_ptr);
				free(file_location);
				fclose(fp);
				return(-1);
			}
		}

		/*
			now append the line to the storelist, except if we were told to skip this one. Then just free the data.
		*/
		if(!skip)
		{
			if(d_list_append(debuglvl, &storelist, line_ptr) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "inserting line into temporary storage list failed (in: %s).", __FUNCTION__);

				/* cleanup */
				d_list_cleanup(debuglvl, &storelist);
				free(line_ptr);
				free(file_location);
				fclose(fp);
				return(-1);
			}
			
			/* we dont need this anymore */
			line_ptr = NULL;
		}
		else
		{
			/* free and null */
			free(line_ptr);
			line_ptr = NULL;
			
			skip = 0;
		}
	}

	/*
		if we are not overwriting and the type of data is already found somewhere, we try to insert is just below
		the last one.
	*/
	if(!overwrite && found)
	{
		if(!(line_ptr = malloc(sizeof(line))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);

			/* cleanup */
			d_list_cleanup(debuglvl, &storelist);
			free(file_location);
			fclose(fp);
			return(-1);
		}
		
		/* assemble the line */
//TODO
		snprintf(line_ptr, sizeof(line), "%s=\"%s\"\n", question, answer);
		
		/*
			loop the list bottom up so we match the last one first
		*/
		for(d_node = storelist.bot; d_node; d_node = d_node->prev)
		{
			if(!(tmp_line_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s)", __FUNCTION__);

				/* cleanup */
				d_list_cleanup(debuglvl, &storelist);
				free(file_location);
				free(line_ptr);
				fclose(fp);
				return(-1);
			}

			/*
				check if the line is the same. If so insert after it.
			*/
			if(strncmp(question, tmp_line_ptr, strlen(question)) == 0)
			{
				if(d_list_insert_after(debuglvl, &storelist, d_node, line_ptr) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "inserting line into temporary storage list failed (in: %s).", __FUNCTION__);
					
					/* cleanup */
					d_list_cleanup(debuglvl, &storelist);
					free(file_location);
					free(line_ptr);
					fclose(fp);
					return(-1);
				}
				
				/* we no longer need these */
				tmp_line_ptr = NULL;
				line_ptr = NULL;
				
				/* after inserting we're done */
				break;
				
			}
		}
	}
	
	/*
		if its not found, we insert it at the end of the list
	*/
	if(found == 0)
	{
		/* first alloc */
		if(!(line_ptr = malloc(sizeof(line))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));

			/* cleanup */
			d_list_cleanup(debuglvl, &storelist);
			free(file_location);
			fclose(fp);
			return(-1);
		}

		/* assemble the line */
//TODO
		snprintf(line_ptr, sizeof(line), "%s=\"%s\"\n", question, answer);

		/* append into the list */
		if(d_list_append(debuglvl, &storelist, line_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "inserting line into temporary storage list failed (in: %s).", __FUNCTION__);

			/* cleanup */
			d_list_cleanup(debuglvl, &storelist);
			free(file_location);
			free(line_ptr);
			fclose(fp);
			return(-1);
		}
		
		/* we no longer need this */
		line_ptr = NULL;
	}

	/*
		close the file
	*/
	if(fclose(fp) < 0)
	{
		(void)vrprint.error(-1, "Error", "closing file '%s' failed: %s.", file_location, strerror(errno));

		/* cleanup */
		d_list_cleanup(debuglvl, &storelist);
		free(file_location);
		return(-1);
	}

	/*
		now open the file for writing
	*/
	if(!(fp = vuurmuur_fopen(file_location, "w+")))
	{
		(void)vrprint.error(-1, "Error", "unable to open file '%s' for writing: %s (in: %s).", file_location, strerror(errno), __FUNCTION__);
		
		/* cleanup */
		d_list_cleanup(debuglvl, &storelist);
		free(file_location);
		return(-1);
	}
	
	/*
		print the list into the file
	*/
	for(d_node = storelist.top; d_node; d_node = d_node->next)
	{
		fprintf(fp, "%s", (char *)d_node->data);
	}

	if(fclose(fp) < 0)
	{
		(void)vrprint.error(-1, "Error", "closing file '%s' failed: %s.", file_location, strerror(errno));

		/* cleanup */
		d_list_cleanup(debuglvl, &storelist);
		free(file_location);
		return(-1);
	}

	/*
		destroy the temp storage
	*/
	d_list_cleanup(debuglvl, &storelist);
	free(file_location);

	return(retval);
}

// opening the backend
static int
open_textdir(int debuglvl, void *backend, int mode, int type)
{
	struct TextdirBackend_ *ptr;

	if(!backend)
		return(-1);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "open_textdir: backend address: %p", backend);

	ptr = (struct TextdirBackend_ *)backend;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "open_textdir: ptr = %p", ptr);

	// see if we like the file permissions of the textdirroot
	if(!(stat_ok(debuglvl, ptr->textdirlocation)))
		return(-1);

	if(ptr->backend_open == 1)
	{
		(void)vrprint.error(-1, "Error", "opening textdir failed: already open.");
		return(-1);
	}
	else
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "open_textdir: setting open to 1");

		// set to open
		ptr->backend_open = 1;
	}

	// now if were opening for type CAT_ZONES, setup the regex
	if(type == CAT_ZONES)
	{
		// regex setup
		if(!(ptr->zonename_reg = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));

			// set the backend to closed again
			ptr->backend_open = 0;

			return(-1);
		}

		// this regex is defined in libvuurmuur -> lib.h
		if(regcomp(ptr->zonename_reg, ZONE_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression failed.");

			// set the backend to closed again
			ptr->backend_open = 0;

			// free and set to NULL, for if we for some reason dont quit after this error (again better safe than sorry, right?)
			free(ptr->zonename_reg);
			ptr->zonename_reg = NULL;

			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "open_textdir: setting up regex for zonename succes.");
	}
	else if(type == CAT_SERVICES)
	{
		// regex setup
		if(!(ptr->servicename_reg = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));

			// set the backend to closed again
			ptr->backend_open = 0;

			return(-1);
		}

		// this regex is defined in libvuurmuur -> lib.h
		if(regcomp(ptr->servicename_reg, SERV_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression failed.");

			// set the backend to closed again
			ptr->backend_open = 0;

			// free and set to NULL, for if we for some reason dont quit after this error (again better safe than sorry, right?)
			free(ptr->servicename_reg);
			ptr->servicename_reg = NULL;

			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "open_textdir: setting up regex for servicename succes.");
	}
	else if(type == CAT_INTERFACES)
	{
		// regex setup
		if(!(ptr->interfacename_reg = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));

			// set the backend to closed again
			ptr->backend_open = 0;

			return(-1);
		}

		// this regex is defined in libvuurmuur -> lib.h
		if(regcomp(ptr->interfacename_reg, IFAC_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression failed.");

			// set the backend to closed again
			ptr->backend_open = 0;

			// free and set to NULL, for if we for some reason dont quit after this error (again better safe than sorry, right?)
			free(ptr->interfacename_reg);
			ptr->interfacename_reg = NULL;

			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "open_textdir: setting up regex for interfacename succes.");
	}

	return(0);
}

static int
close_textdir(int debuglvl, void *backend, int type)
{
	struct TextdirBackend_ *ptr;

	// safetybelt
	if(!backend)
		return(-1);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "close_textdir: backend: %p", backend);

	ptr = (struct TextdirBackend_ *)backend;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "close_textdir: ptr = %p", ptr);

	if(ptr->backend_open)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "close_textdir: closing: setting backend_open to 0");

		ptr->backend_open = 0;
	}
	else
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "close_textdir: already closed");
	}

	// cleanup regex
	if(type == CAT_ZONES && ptr->zonename_reg)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "close_textdir: cleaning up regex.");

		regfree(ptr->zonename_reg);
		free(ptr->zonename_reg);
		ptr->zonename_reg = NULL;
	}
	else if(type == CAT_SERVICES && ptr->servicename_reg)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "close_textdir: cleaning up regex.");

		regfree(ptr->servicename_reg);
		free(ptr->servicename_reg);
		ptr->servicename_reg = NULL;
	}
	else if(type == CAT_INTERFACES && ptr->interfacename_reg)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "close_textdir: cleaning up regex.");

		regfree(ptr->interfacename_reg);
		free(ptr->interfacename_reg);
		ptr->interfacename_reg = NULL;
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "close_textdir: end");

	return(0);
}

/*	list_textdir

	Listing the items in the backend. A number of checks are done:
		- regex validation
		- stat (file perms) validation
		- .conf suffix check (optional)

	Returns a pointer to the name, or NULL when done.
*/
static char
*list_textdir(int debuglvl, void *backend, char *name, int *zonetype, int type)
{
	struct TextdirBackend_ 		*ptr = NULL;
	char 				dir_location[512],
					netdir_location[512],
					hostdir_location[512],
					groupdir_location[512],
					*file_location;
	char				cur_zonename[MAX_HOST_NET_ZONE] = "";
	struct dirent			*dir_entry_p;
	int				done = 0;


	/* safety */
	if(!backend || !name || !zonetype)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}


	/* check if the backend is opened */
	ptr = (struct TextdirBackend_ *)backend;
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "Backend not opened yet (in: list_textdir).");
		return(NULL);
	}


	/* services */
	if(type == CAT_SERVICES)
	{
		// set the dir location
		snprintf(dir_location, sizeof(dir_location), "%s/services", ptr->textdirlocation);

		while(!done)
		{
			// check if already open
			if(!ptr->service_p)
			{
				// open the dir
				if(!(ptr->service_p = vuurmuur_opendir(dir_location)))
				{
					(void)vrprint.error(-1, "Error", "unable to open '%s', %s (in: list_textdir, opendir).", dir_location, strerror(errno));
					return(NULL);
				}
			}

			// now read the contents of the dir
			if((dir_entry_p = readdir(ptr->service_p)) != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: name: %s.", dir_entry_p->d_name);

				strlcpy(name, dir_entry_p->d_name, MAX_SERVICE);

				// only return names which do not start with a dot (hidden files) and don't open . and ..
				if((strncmp(name, ".", 1) != 0) && (strcmp(name, "..") != 0))
				{
					// now validate the name
					if(validate_servicename(debuglvl, name, ptr->servicename_reg) == 0)
					{
						// determine the location of the file
						if(!(file_location = get_filelocation(debuglvl, backend, name, TYPE_SERVICE)))
							return(NULL);
						
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: service '%s', file: '%s'.", name, file_location);

						// now stat it
						if(stat_ok(debuglvl, file_location))
						{
							free(file_location);

							*zonetype = TYPE_SERVICE;
							return(name);
						}

						free(file_location);
					}
				}
			}
			// close the dir if we cant read no more
			else
			{
				// close the dir
				done = 1;
				closedir(ptr->service_p);
				ptr->service_p = NULL;
			}
		}
	}
	// interfaces
	else if(type == CAT_INTERFACES)
	{
		// set the dirlocation
		snprintf(dir_location, sizeof(dir_location), "%s/interfaces", ptr->textdirlocation);

		while(done == 0)
		{
			// check if already open, if not: open it
			if(!ptr->interface_p)
			{
				// open the dir
				if(!(ptr->interface_p = vuurmuur_opendir(dir_location)))
				{
					(void)vrprint.error(-1, "Error", "Unable to open '%s', %s.", dir_location, strerror(errno));
					return(NULL);
				}
			}

			if((dir_entry_p = readdir(ptr->interface_p)) != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: name: '%s'.", dir_entry_p->d_name);

				if((strncmp(dir_entry_p->d_name, ".", 1) != 0) && (strcmp(dir_entry_p->d_name, "..") != 0) && (strlen(dir_entry_p->d_name) > 5))
				{
					// this way we check if filename ends with '.conf'
					if(	dir_entry_p->d_name[strlen(dir_entry_p->d_name)-5] == '.' &&
						dir_entry_p->d_name[strlen(dir_entry_p->d_name)-4] == 'c' &&
						dir_entry_p->d_name[strlen(dir_entry_p->d_name)-3] == 'o' &&
						dir_entry_p->d_name[strlen(dir_entry_p->d_name)-2] == 'n' &&
						dir_entry_p->d_name[strlen(dir_entry_p->d_name)-1] == 'f'
					)
					{
						// make sure we dont allow too long filenames
						if((strlen(dir_entry_p->d_name) < MAX_INTERFACE+5))
						{
							strlcpy(ptr->interface, dir_entry_p->d_name, (strlen(dir_entry_p->d_name)-5)+1);
							ptr->interface[strlen(dir_entry_p->d_name)-5]='\0';

							if(validate_interfacename(debuglvl, ptr->interface, ptr->interfacename_reg) == 0)
							{
								*zonetype = TYPE_INTERFACE;
								strlcpy(name, ptr->interface, MAX_INTERFACE);

								// determine the location of the file
								if(!(file_location = get_filelocation(debuglvl, backend, name, TYPE_INTERFACE)))
									return(NULL);
								
								if(debuglvl >= HIGH)
									(void)vrprint.debug("Debug", "list_textdir: interface '%s', file: '%s'.", name, file_location);

								// now stat it
								if(stat_ok(debuglvl, file_location))
								{
									free(file_location);

									if(debuglvl >= HIGH)
										(void)vrprint.debug("Debug", "list_textdir: returning name: '%s'.", name);
									
									return(name);
								}

								// free temp mem
								free(file_location);
							}
						}
						else
						{
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "list_textdir: '%s' is too long.", dir_entry_p->d_name);
						}
					}
				}
			}
			else
			{
				done=1;
				closedir(ptr->interface_p);
				ptr->interface_p = NULL;
			}
		}
	}
	else if(type == CAT_ZONES)
	{
		while(!done)
		{
			// this is the base dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones", ptr->textdirlocation);

			if(ptr->host_p != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: getting a hostname.");


				if((dir_entry_p = readdir(ptr->host_p)) != NULL)
				{
					// make sure that we dont use '.' and '..' and files that are not long enough to be serious
					if((strncmp(dir_entry_p->d_name, ".", 1) != 0) && (strcmp(dir_entry_p->d_name, "..") != 0) && (strlen(dir_entry_p->d_name) > 5))
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: host name: %s.", dir_entry_p->d_name);

						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: item: %s, %d.", dir_entry_p->d_name, strlen(dir_entry_p->d_name));

						// max sure the name is not too long
						if(strlen(dir_entry_p->d_name) < MAX_HOST + 5)
						{
							strlcpy(ptr->cur_host, dir_entry_p->d_name, (strlen(dir_entry_p->d_name)-5)+1);
							ptr->cur_host[strlen(dir_entry_p->d_name)-5]='\0';

							snprintf(cur_zonename, sizeof(cur_zonename), "%s.%s.%s", ptr->cur_host, ptr->cur_network, ptr->cur_zone);

							// lets check against regex
							if(validate_zonename(debuglvl, cur_zonename, 1, NULL, NULL, NULL, ptr->zonename_reg) == 0)
							{
								// determine the location of the file
								if(!(file_location = get_filelocation(debuglvl, backend, cur_zonename, TYPE_HOST)))
									return(NULL);
								
								if(debuglvl >= HIGH)
									(void)vrprint.debug("Debug", "list_textdir: host '%s', file: '%s'.", cur_zonename, file_location);

								// now stat it
								if(stat_ok(debuglvl, file_location))
								{
									free(file_location);

									*zonetype = TYPE_HOST;

									strlcpy(name, cur_zonename, MAX_HOST_NET_ZONE);
									return(name);
								}

								free(file_location);
							}
						}
					}
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: host dir closed.");
					
					closedir(ptr->host_p);
					ptr->host_p = NULL;
				}
			}
			else if(ptr->group_p != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: getting a groupname.");

				if((dir_entry_p = readdir(ptr->group_p)) != NULL)
				{
					// make sure that we dont use '.' and '..' and files that are not long enough to be serious
					if((strncmp(dir_entry_p->d_name, ".", 1) != 0) && (strcmp(dir_entry_p->d_name, "..") != 0) && (strlen(dir_entry_p->d_name) > 5))
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: group name: %s.", dir_entry_p->d_name);

						// max sure the name is not too long
						if(strlen(dir_entry_p->d_name) < MAX_HOST + 5)
						{
							strlcpy(ptr->cur_host, dir_entry_p->d_name, (strlen(dir_entry_p->d_name)-6)+1);
							ptr->cur_host[strlen(dir_entry_p->d_name)-6]='\0';

							snprintf(cur_zonename, sizeof(cur_zonename), "%s.%s.%s", ptr->cur_host, ptr->cur_network, ptr->cur_zone);

							// lets check against regex
							if(validate_zonename(debuglvl, cur_zonename, 1, NULL, NULL, NULL, ptr->zonename_reg) == 0)
							{
								// determine the location of the file
								if(!(file_location = get_filelocation(debuglvl, backend, cur_zonename, TYPE_GROUP)))
									return(NULL);
								
								if(debuglvl >= HIGH)
									(void)vrprint.debug("Debug", "list_textdir: group '%s', file: '%s'.", cur_zonename, file_location);

								// now stat it
								if(stat_ok(debuglvl, file_location))
								{
									free(file_location);

									*zonetype = TYPE_GROUP;

									strlcpy(name, cur_zonename, MAX_HOST_NET_ZONE);
									return(name);
								}

								free(file_location);
							}
						}
					}
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: group dir closed.");

					closedir(ptr->group_p);
					ptr->group_p = NULL;
				}
			}
			else if(ptr->network_p != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: getting a networkname.");

				if((dir_entry_p = readdir(ptr->network_p)) != NULL)
				{
					snprintf(netdir_location, sizeof(netdir_location), "%s/zones/%s/networks", ptr->textdirlocation, ptr->cur_zone);

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: network entry: %s.", dir_entry_p->d_name);

					if((strncmp(dir_entry_p->d_name, ".", 1) != 0) && (strcmp(dir_entry_p->d_name, "..") != 0))
					{
						strlcpy(ptr->cur_network, dir_entry_p->d_name, MAX_NETWORK);
						if(debuglvl >= HIGH) vrprint.debug("Debug", "list_textdir: network name: %s.", dir_entry_p->d_name);

						// open the hostdir
						snprintf(hostdir_location, sizeof(hostdir_location), "%s/%s/hosts", netdir_location, dir_entry_p->d_name);
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: opening host dir: %s.", hostdir_location);

						// this is allowed to fail, if it does, is will be NULL, and will be detected above
						ptr->host_p = vuurmuur_opendir(hostdir_location);

						// open the groupdir
						snprintf(groupdir_location, sizeof(groupdir_location), "%s/%s/groups", netdir_location, dir_entry_p->d_name);
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: opening group dir: %s.", groupdir_location);

						// this is allowed to fail, if it does, is will be NULL, and will be detected above
						ptr->group_p = vuurmuur_opendir(groupdir_location);

						snprintf(cur_zonename, sizeof(cur_zonename), "%s.%s", ptr->cur_network, ptr->cur_zone);

						// lets check against regex
						if(validate_zonename(debuglvl, cur_zonename, 1, NULL, NULL, NULL, ptr->zonename_reg) == 0)
						{
							// determine the location of the file
							if(!(file_location = get_filelocation(debuglvl, backend, cur_zonename, TYPE_NETWORK)))
								return(NULL);
							
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "list_textdir: network '%s', file: '%s'.", cur_zonename, file_location);

							// now stat it
							if(stat_ok(debuglvl, file_location))
							{
								free(file_location);

								if(debuglvl >= HIGH)
									(void)vrprint.debug("Debug", "list_textdir: '%s' ('%s', '%s').", cur_zonename, ptr->cur_network, ptr->cur_zone);

								*zonetype = TYPE_NETWORK;

								strlcpy(name, cur_zonename, MAX_NET_ZONE);
								return(name);
							}

							free(file_location);
						}
					}
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: network dir closed.");

					closedir(ptr->network_p);
					ptr->network_p = NULL;
				}
			}
			else if(ptr->zone_p != NULL)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: getting a zonename.");

				if((dir_entry_p = readdir(ptr->zone_p)) != NULL)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: zone entry: %s.", dir_entry_p->d_name);

					if((strncmp(dir_entry_p->d_name, ".", 1) != 0) && (strcmp(dir_entry_p->d_name, "..") != 0))
					{
						strlcpy(ptr->cur_zone, dir_entry_p->d_name, MAX_ZONE);

						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: zone name: %s.", dir_entry_p->d_name);

						// open the networkdir
						snprintf(netdir_location, sizeof(netdir_location), "%s/%s/networks", dir_location, dir_entry_p->d_name);
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "list_textdir: opening: %s.", netdir_location);

						// this is allowed to fail, if it does, is will be NULL, and will be detected above
						ptr->network_p = vuurmuur_opendir(netdir_location);

						// lets check against regex
						if(validate_zonename(debuglvl, dir_entry_p->d_name, 1, NULL, NULL, NULL, ptr->zonename_reg) == 0)
						{
							// determine the location of the file
							if(!(file_location = get_filelocation(debuglvl, backend, dir_entry_p->d_name, TYPE_ZONE)))
								return(NULL);
							
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "list_textdir: zone '%s', file: '%s'.", ptr->cur_zone, file_location);

							// now stat it
							if(stat_ok(debuglvl, file_location))
							{
								free(file_location);

								if(debuglvl >= HIGH)
									(void)vrprint.debug("Debug", "list_textdir: zone '%s'.", ptr->cur_zone);

								*zonetype = TYPE_ZONE;

								strlcpy(name, dir_entry_p->d_name, MAX_ZONE);
								return(name);
							}

							free(file_location);
						}
					}
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "list_textdir: zone dir closing.");

					closedir(ptr->zone_p);
					ptr->zone_p = NULL;

					done = 1;
				}
			}

			if(!done && !ptr->zone_p)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "list_textdir: opening the zonesdir.");

				// open the dir
				if(!(ptr->zone_p = vuurmuur_opendir(dir_location)))
				{
					(void)vrprint.error(-1, "Error", "unable to open directory: %s: %s.", dir_location, strerror(errno));
					return(NULL);
				}
			}
		}
	}
	else
	{
		(void)vrprint.error(-1, "Internal Error", "unknown type '%d'.", type);
		return(NULL);
	}

	return(NULL);
}

// setting up the backend for first use
static int init_textdir(int debuglvl, void *backend, int type)
{
//TODO
	return(0);
}

/*	add item to the backend

*/
static int
add_textdir(int debuglvl, void *backend, char *name, int type)
{
	FILE			*fp;
	struct TextdirBackend_	*ptr;
	char			*file_location,
				dir_location[512],
				hostname[MAX_HOST],
				networkname[MAX_NETWORK],
				zonename[MAX_ZONE];
	int			fd;

	/* safety */
	if(!backend || !name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if the backend is open */
	ptr = (struct TextdirBackend_ *)backend;
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "Backend not opened yet (in: add_textdir).");
		return(-1);
	}

	/* determine the location of the file */
	if(!(file_location = get_filelocation(debuglvl, backend, name, type)))
		return(-1);

	/* check if the file already exist */
	if((fp = fopen(file_location, "r")))
	{
		(void)vrprint.error(-1, "Error", "creating %s failed: file exists.", file_location);

		fclose(fp);
		free(file_location);

		return(-1);
	}

	// create the dirs for zones and networks
	if(type == TYPE_ZONE || type == TYPE_NETWORK)
	{
		// split up the name
		if(validate_zonename(debuglvl, name, 0, zonename, networkname, hostname, ptr->zonename_reg) != 0)
		{
			(void)vrprint.error(-1, "Error", "Zonename '%s' is not valid.", name);
			return(-1);
		}

		if(type == TYPE_ZONE)
		{
			// zone dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s", ptr->textdirlocation, zonename);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "Creating directory %s failed: %s.", dir_location, strerror(errno));
				return(-1);
			}

			// network dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks", ptr->textdirlocation, zonename);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "Creating directory %s failed: %s.", dir_location, strerror(errno));
				return(-1);
			}
		}
		else if(type == TYPE_NETWORK)
		{
			// network dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s", ptr->textdirlocation, zonename, networkname);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "Creating directory %s failed: %s.", dir_location, strerror(errno));
				return(-1);
			}

			// host dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s/hosts", ptr->textdirlocation, zonename, networkname);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "Creating directory %s failed: %s.", dir_location, strerror(errno));
				return(-1);
			}

			// group dir
			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s/groups", ptr->textdirlocation, zonename, networkname);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "Creating directory %s failed: %s.", dir_location, strerror(errno));
				return(-1);
			}
		}
	}

	// now open for writing (file will be created)
	if((fd = open(file_location, O_WRONLY|O_CREAT|O_EXCL, 0600)) == -1)
	{
		(void)vrprint.error(-1, "Error", "Creating %s failed: %s (in: add_textdir).", file_location, strerror(errno));

		free(file_location);

		return(-1);
	}

	if(write(fd, "ACTIVE=\"\"\n", 10) == -1)
	{
		(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
		return(-1);
	}

	if(type == TYPE_HOST)
	{
		if(write(fd, "IPADDRESS=\"\"\n", 13) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "MAC=\"\"\n", 7) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}
	}

	if(type == TYPE_GROUP)
	{
		if(write(fd, "MEMBER=\"\"\n", 10) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}
	}

	if(type == TYPE_NETWORK)
	{
		if(write(fd, "NETWORK=\"\"\n", 11) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "NETMASK=\"\"\n", 11) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "INTERFACE=\"\"\n", 13) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}
	}

	if(type == TYPE_SERVICE)
	{
		/*
		if(write(fd, "PROTO=\"\"\n", 9) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
		}
		*/

		if(write(fd, "TCP=\"\"\n", 7) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "UDP=\"\"\n", 7) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "ICMP=\"\"\n", 8) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "GRE=\"\"\n", 7) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "BROADCAST=\"\"\n", 13) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

	}

	if(type == TYPE_INTERFACE)
	{
		if(write(fd, "IPADDRESS=\"\"\n", 13) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "INTERFACE=\"\"\n", 13) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}

		if(write(fd, "VIRTUAL=\"\"\n", 11) == -1)
		{
			(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
			return(-1);
		}
	}

	if(write(fd, "COMMENT=\"\"\n", 11) == -1)
	{
		(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
		return(-1);
	}

/*
	if(write(fd, "\n# end of file\n", 15) == -1)
	{
		(void)vrprint.error(-1, "Error", "write: %s", strerror(errno));
		return(-1);
	}
*/
	fsync(fd);

	if(close(fd) == -1)
	{
		(void)vrprint.error(-1, "Error", "closing file descriptor failed.");

		free(file_location);

		return(-1);
	}

	return(0);
}

/*	del_textdir

	Delete from the textdir.
	
	Returncodes:
		0: ok
		-1: error
*/
static int
del_textdir(int debuglvl, void *backend, char *name, int type, int recurs)
{
	char			*file_location,
				dir_location[512],
				hostname[MAX_HOST],
				networkname[MAX_NETWORK],
				zonename[MAX_ZONE];
	struct TextdirBackend_	*ptr;

	/* safety */
	if(!backend || !name)
	{
		(void)vrprint.error(-1, "Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if the backend was properly openend */
	ptr = (struct TextdirBackend_ *)backend;
	if(!ptr->backend_open)
	{
		(void)vrprint.error(-1, "Error", "backend not opened yet (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* determine the location of the file */
	if(!(file_location = get_filelocation(debuglvl, backend, name, type)))
		return(-1);

	/* see if we like the file permissions */
	if(!(stat_ok(debuglvl, file_location)))
		return(-1);

	/* name splitting only needed for network and zone, as host and group just use the file_location
	   this is because network and zone need to remove directories as well
	*/
	if(type == TYPE_ZONE || type == TYPE_NETWORK)
	{
		// split up the name
		if(validate_zonename(debuglvl, name, 0, zonename, networkname, hostname, ptr->zonename_reg) != 0)
		{
			(void)vrprint.error(-1, "Error", "Zonename '%s' is not valid.", name);
			return(-1);
		}
	}

	/*
		HERE WE DO THE REMOVAL
	*/

	if(type == TYPE_HOST)
	{
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "Deleting host file for '%s': %s.", name, strerror(errno));

			free(file_location);

			return(-1);
		}

		(void)vrprint.info("Info", "host '%s' deleted from disk.", name);
	}
	else if(type == TYPE_GROUP)
	{
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting group file for '%s': %s.", name, strerror(errno));

			free(file_location);

			return(-1);
		}

		(void)vrprint.info("Info", "group '%s' deleted from disk.", name);
	}
	else if(type == TYPE_NETWORK)
	{
		// first check the hosts dir
		snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s/hosts", ptr->textdirlocation, zonename, networkname);

		if(rmdir(dir_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting host dir failed: %s", strerror(errno));

			free(file_location);

			return(-1);
		}

		// second check the group dir
		snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s/groups", ptr->textdirlocation, zonename, networkname);
		if(rmdir(dir_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting group dir '%s' failed: %s.", dir_location, strerror(errno));

			snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s/hosts", ptr->textdirlocation, zonename, networkname);
			if(mkdir(dir_location, 0700) < 0)
			{
				(void)vrprint.error(-1, "Error", "recreating hostdir failed: %s.", strerror(errno));

				free(file_location);

				return(-1);
			}

			// now quit
			free(file_location);

			return(-1);
		}

		// the network.config file
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting network.config: %s.", strerror(errno));

			free(file_location);

			return(-1);
		}

		// the network dir
		snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks/%s", ptr->textdirlocation, zonename, networkname);
		if(rmdir(dir_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting network dir: %s.", strerror(errno));

			free(file_location);

			return(-1);
		}

		(void)vrprint.info("Info", "Network '%s' deleted from disk.", name);
	}
	else if(type == TYPE_ZONE)
	{
		// first check the network
		snprintf(dir_location, sizeof(dir_location), "%s/zones/%s/networks", ptr->textdirlocation, name);
		if(rmdir(dir_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting network dir failed: %s", strerror(errno));

			free(file_location);

			return(-1);
		}

		// the zone.config file
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting zone.config failed: %s.", strerror(errno));

			free(file_location);

			return(-1);
		}

		// first check the hosts dir
		snprintf(dir_location, sizeof(dir_location), "%s/zones/%s", ptr->textdirlocation, name);
		if(rmdir(dir_location) < 0)
		{
			(void)vrprint.error(-2, "Error", "deleting zone dir failed: %s", strerror(errno));

			free(file_location);

			return(-1);
		}
	}
	else if(type == TYPE_SERVICE)
	{
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting service %s: %s (%s).", name, strerror(errno), file_location);

			free(file_location);

			return(-1);
		}
	}
	else if(type == TYPE_INTERFACE)
	{
		if(remove(file_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "deleting interface %s: %s (%s).", name, strerror(errno), file_location);

			free(file_location);

			return(-1);
		}
	}
	// handle unknown
	else
	{
		(void)vrprint.error(-1, "Internal Error", "unknown type: %d (in: del_textdir).", type);
		free(file_location);
		return(-1);
	}

	// cleanup
	free(file_location);

	return(0);
}

/*	conf_textdir

	Loads the config settings from the plugin config file.

	Returncodes:
		 0: ok
		-1: error
*/
static int conf_textdir(const int debuglvl, void *backend)
{
	int			retval = 0,
				result = 0;
	char			configfile_location[512];
	struct TextdirBackend_	*ptr = NULL;

	/* safety first */
	if(!backend)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(!(ptr = (struct TextdirBackend_ *)backend))
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* assemble config location */
	if(snprintf(configfile_location, sizeof(configfile_location), "%s/textdir.conf", conf.plugindir) >= sizeof(configfile_location))
	{
		(void)vrprint.error(-1, "Internal Error", "could not determine configfile location: locationstring overflow (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* now get the backend location from the configfile */
	result=ask_configfile(debuglvl, "LOCATION", ptr->textdirlocation, configfile_location, sizeof(ptr->textdirlocation));
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "failed to get the textdir-root from: %s. Please make sure LOCATION is set (in: %s).", configfile_location, __FUNCTION__);
		retval = -1;
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "no information about the location of the backend in '%s' (in: %s).", configfile_location, __FUNCTION__);
		retval = -1;
	}
	else
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "conf_textdir: Textdir location: LOCATION = %s.", ptr->textdirlocation);
	}

	return(retval);
}

static int
setup_textdir(int debuglvl, void **backend)
{
	struct TextdirBackend_ *ptr;

	if(!(ptr = malloc(sizeof(struct TextdirBackend_))))
	{
		(void)vrprint.error(-1, "Error", "malloc (in: _init (textdir.c)).");
		return(-1);
	}

	// set the versions
	ptr->version_major = TEXTDIRVERSION_MAJOR;
	ptr->version_minor = TEXTDIRVERSION_MINOR;
	ptr->version_sub = TEXTDIRVERSION_SUB;

	// start closed of course
	ptr->backend_open = 0;

	// not used yet
	ptr->writable = 0; // todo

	ptr->zone_p = NULL;
	ptr->network_p = NULL;
	ptr->host_p = NULL;
	ptr->group_p = NULL;
	ptr->service_p = NULL;
	ptr->interface_p = NULL;
	
	ptr->file = NULL;

	//
	ptr->zonename_reg = NULL;
	ptr->servicename_reg = NULL;
	ptr->interfacename_reg = NULL;

	// return the backend pointer to the caller
	*backend = (void *)ptr;

	return(0);
}

//TODO:
//	set version
//	set rw/ro
void _init(void)
{
	BackendFunctions.ask = ask_textdir;
	BackendFunctions.tell = tell_textdir;
	BackendFunctions.open = open_textdir;
	BackendFunctions.close = close_textdir;
	BackendFunctions.list = list_textdir;
	BackendFunctions.init = init_textdir;
	BackendFunctions.add = add_textdir;
	BackendFunctions.del = del_textdir;
	BackendFunctions.conf = conf_textdir;
	BackendFunctions.setup = setup_textdir;
}

void _fini(void)
{
	//fprintf(stdout, "textdir fini start (%d)\n", sizeof(struct TextdirBackend_));
	//fprintf(stdout, "textdir fini done\n");
}

